"""Produces reporting to help interpret analysis and campaigns."""
import logging

import matplotlib.pyplot as plt  # type: ignore
import pandas as pd
import seaborn as sns

from pricer import config as cfg, io

logger = logging.getLogger(__name__)
sns.set(rc={"figure.figsize": (3, 3)})


def have_in_bag() -> str:
    """Prints expected profits, make sure its in your bag."""
    sell_policy = io.reader("outputs", "sell_policy", "parquet")
    sell_policy = sell_policy.set_index("item")

    sell_policy = sell_policy[
        sell_policy["sell_estimated_profit"] > sell_policy["profit_feasible"]
    ]

    print(sell_policy["sell_estimated_profit"])
    return sell_policy[["sell_estimated_profit"]].astype(int).to_html()


def make_missing() -> str:
    """Prints details of items unable to be made."""
    make_policy = io.reader("outputs", "make_policy", "parquet")

    make_me = make_policy[
        (make_policy["user_make_pass"] == 0) & (make_policy["make_actual"] > 0)
    ]["make_actual"]
    make_me.name = "Automake"
    print(make_me)

    make_main = make_policy[
        (make_policy["user_make_pass"] == 1) & (make_policy["make_ideal"] > 0)
    ]["make_ideal"]
    make_main.name = "Make on main"
    print(make_main)

    make_should = make_policy[
        (
            (make_policy["user_make_pass"] == 0)
            & (make_policy["make_ideal"] > make_policy["make_actual"])
            & (make_policy["user_Make"] == 1)
        )
    ]["make_ideal"]
    make_should.name = "Missing mats"
    print(make_should)

    making_html = pd.DataFrame(index=pd.concat([make_me, make_main, make_should]).index)
    making_html = (
        making_html.join(make_me)
        .join(make_main)
        .join(make_should)
        .fillna(0)
        .astype(int)
    )

    return making_html.to_html()


def produce_item_reporting() -> None:
    """Collate item information and prepare feasibility chart."""
    item_table = io.reader("intermediate", "item_table", "parquet")
    buy_policy = io.reader("outputs", "buy_policy", "parquet").set_index("item")
    sell_policy = io.reader("outputs", "sell_policy", "parquet").set_index("item")
    make_policy = io.reader("outputs", "make_policy", "parquet")

    item_info = (
        item_table.join(buy_policy[[x for x in buy_policy if x not in item_table]])
        .join(sell_policy[[x for x in sell_policy if x not in item_table]])
        .join(make_policy[[x for x in make_policy if x not in item_table]])
    )

    item_info = item_info[sorted(item_info.columns)]

    # item_reporting = {
    #     item: pd.DataFrame(item_info.loc[item]).to_html() for item in item_info.index
    # }
    io.writer(item_info, "reporting", "item_info", "parquet")

    listing_profits = io.reader("reporting", "listing_profits", "parquet")

    MAX_LISTINGS = cfg.us["analysis"]["MAX_LISTINGS_PROBABILITY"]
    for item in listing_profits.columns:
        plt.figure()
        listing_profits[item].plot(title=f"List profit {item}")
        pd.Series([sell_policy.loc[item, "profit_feasible"]] * MAX_LISTINGS).plot()
        plt.savefig(f"data/reporting/feasible/{item}.png")
        plt.close()


def produce_listing_items() -> None:
    """Generte the item listing on current AH."""
    listing_each = io.reader("intermediate", "listing_each", "parquet")
    item_info = io.reader("reporting", "item_info", "parquet")

    for item in cfg.ui:
        plt.figure()
        list_item = listing_each[
            (listing_each["item"] == item) & (listing_each["list_price_z"] < 10)
        ]
        list_item = list_item["list_price_per"].sort_values().reset_index(drop=True)
        list_item.plot(title=f"Current AH listings {item}")

        pd.Series(
            [item_info.loc[item, "material_make_cost"]] * list_item.shape[0]
        ).plot()

        plt.savefig(f"data/reporting/listing_item/{item}.png")
        plt.close()


def produce_activity_tracking() -> None:
    """Produce chart of item prices, sold and bought for."""
    bean_results = io.reader("cleaned", "bean_results", "parquet")
    bean_results["date"] = bean_results["timestamp"].dt.date.astype("datetime64")
    bean_sales = bean_results.groupby(["item", "date"])["buyout_per"].mean()
    bean_sales.name = "sell_price"

    bean_purchases = io.reader("cleaned", "bean_purchases", "parquet")
    bean_purchases["date"] = bean_purchases["timestamp"].dt.date.astype("datetime64")
    bean_buys = bean_purchases.groupby(["item", "date"])["buyout_per"].mean()
    bean_buys.name = "buy_price"

    bb_history = io.reader("cleaned", "bb_history", "parquet")
    bb_history = bb_history[bb_history["date"] >= bean_results["date"].min()]
    bb_history = bb_history.set_index(["item", "date"])

    activity = bb_history.join(bean_buys).join(bean_sales)
    cols = ["silveravg", "buy_price", "sell_price"]

    for item in cfg.ui:
        if item in activity.index:
            plt.figure()
            activity.loc[item][cols].plot(title=f"Historic activity {item}")
            plt.savefig(f"data/reporting/activity/{item}.png")
            plt.close()


def profit_per_item() -> str:
    """Profits per item as HTML."""
    profits = io.reader("reporting", "profits", "parquet")

    item_profits = (
        (profits.groupby("item")[["total_profit"]].sum() / 10000)
        .astype(int)
        .sort_values(by="total_profit", ascending=False)
    )
    return item_profits.to_html()


def draw_profit_charts() -> None:
    """Create charts of alltime and individual item profits."""
    profits = io.reader("reporting", "profits", "parquet")

    alltime_profit = (
        profits.reset_index().groupby("date")["total_profit"].sum().cumsum() / 10000
    )

    tot = int(alltime_profit[-1])
    daily = int(tot / alltime_profit.shape[0])

    plt.figure()
    alltime_profit.plot(
        title=f"Total profit over all items ({tot} gold, {daily} per day)"
    )
    plt.savefig("data/reporting/profit/_alltime_profits.png")
    plt.close()

    for item in cfg.ui:
        if item in profits.index:
            plt.figure()
            (profits.loc[item, "total_profit"].cumsum() / 10000).plot(
                title=f"Profit {item}"
            )
            plt.savefig(f"data/reporting/profit/{item}.png")
            plt.close()



"""Pages for webserver."""
import logging
from pathlib import Path
from typing import Any, Dict, List

from flask import Flask, redirect, render_template, send_from_directory, url_for

from .. import config as cfg, io, reporting, run, sources


logger = logging.getLogger(__name__)
app = Flask(__name__)
if app.root_path is None:
    raise
app.config["data_path"] = Path(app.root_path).parents[2].joinpath("data")

try:
    item_icon_manifest = io.reader("item_icons", "_manifest", "json")

except FileNotFoundError:
    logger.exception("Reporting files not present, unable to start webserver")


@app.context_processor
def g() -> Dict[Any, Any]:
    """Globals for general purpose."""

    def item_profits() -> str:
        return reporting.have_in_bag()

    def make_missing() -> str:
        return reporting.make_missing()

    def user_items() -> List[str]:
        return sorted(cfg.ui.keys())

    def profit_per_item() -> str:
        return reporting.profit_per_item()

    return dict(
        item_profits=item_profits,
        user_items=user_items,
        make_missing=make_missing,
        profit_per_item=profit_per_item,
    )


@app.route("/")
def home() -> Any:
    """Return homepage."""
    return render_template("home.html")


@app.route("/<path:item_name>")
def item_report(item_name: str) -> Any:
    """Return info on an item."""
    item_info = io.reader("reporting", "item_info", "parquet")
    item_report = item_info.loc[item_name].to_dict()
    return render_template(
        "item_reporting.html", item_name=item_name, item_report=item_report
    )


@app.route("/data_static/item_icons/<path:filename>")
def item_icons(filename: str) -> Any:
    """Returns image icon for items."""
    icon = item_icon_manifest.get(filename, "inv_scroll_03") + ".jpg"
    return send_from_directory(
        Path(app.config["data_path"]).joinpath("item_icons"), icon
    )


@app.route("/data_static/item_plot_profit/<path:item_name>")
def item_plot_profit(item_name: str) -> Any:
    """Returns profit plot for items."""
    path = Path(app.config["data_path"]).joinpath("reporting", "feasible")
    return send_from_directory(path, item_name + ".png")


@app.route("/data_static/item_listing_plot/<path:item_name>")
def item_listing_plot(item_name: str) -> Any:
    """Returns profit plot for items."""
    path = Path(app.config["data_path"]).joinpath("reporting", "listing_item")
    return send_from_directory(path, item_name + ".png")


@app.route("/data_static/item_activity_plot/<path:item_name>")
def item_activity_plot(item_name: str) -> Any:
    """Returns profit plot for items."""
    path = Path(app.config["data_path"]).joinpath("reporting", "activity")
    return send_from_directory(path, item_name + ".png")


@app.route("/data_static/item_profit_plot/<path:item_name>")
def item_profit_plot(item_name: str) -> Any:
    """Returns profit plot for items."""
    path = Path(app.config["data_path"]).joinpath("reporting", "profit")
    return send_from_directory(path, item_name + ".png")


@app.route("/trigger_booty_bay")
def trigger_booty_bay() -> Any:
    """Return homepage."""
    sources.get_bb_data()
    return redirect(url_for("home"))


@app.route("/run_analytics")
def run_analytics() -> Any:
    """Return homepage."""
    run.run_analytics()
    run.run_reporting()
    return redirect(url_for("home"))



"""Collates information to form buy/sell campaigns."""
import logging
from typing import Any, Dict, Tuple

import pandas as pd
from scipy.stats import norm
from slpp import slpp as lua

from pricer import config as cfg, io, utils

logger = logging.getLogger(__name__)


def analyse_buy_policy(MAX_BUY_STD: int = 2) -> None:
    """Create buy policy."""
    logger.debug(f"max buy std {MAX_BUY_STD}")

    item_table = io.reader("intermediate", "item_table", "parquet")

    buy_policy = item_table[item_table["user_Buy"] == True]
    subset_cols = [
        "bbpred_price",
        "bbpred_std",
        "inv_total_all",
        "replenish_qty",
        "user_std_holding",
        "replenish_z",
    ]
    buy_policy = buy_policy[subset_cols]

    listing_each = io.reader("intermediate", "listing_each", "parquet")

    listing_each = listing_each.sort_values("list_price_per")

    rank_list = listing_each.join(buy_policy, on="item").dropna()

    rank_list["sell_rank"] = rank_list.groupby("item")["list_price_per"].rank(
        method="max"
    )

    rank_list = rank_list.drop_duplicates()
    rank_list["updated_rank"] = rank_list["replenish_qty"] - rank_list["sell_rank"]
    rank_list["updated_replenish_z"] = (
        rank_list["updated_rank"] / rank_list["user_std_holding"]
    )

    rank_list["updated_replenish_z"] = rank_list["updated_replenish_z"].clip(
        upper=MAX_BUY_STD
    )

    rank_list = rank_list[rank_list["updated_replenish_z"] > rank_list["list_price_z"]]
    io.writer(rank_list, "reporting", "buy_rank", "parquet")

    buy_policy["buy_price_cap"] = rank_list.groupby("item")["list_price_per"].max()
    buy_policy["buy_price_cap"] = buy_policy["buy_price_cap"].fillna(1).astype(int)

    buy_policy.index.name = "item"
    buy_policy = buy_policy.reset_index()
    io.writer(buy_policy, "outputs", "buy_policy", "parquet")


def encode_buy_campaign(buy_policy: pd.DataFrame) -> Dict[str, Dict[str, Any]]:
    """Encodes buy campaign dataframe into dictionary."""
    cols = ["item", "buy_price_cap"]
    assert (buy_policy.columns == cols).all(), "Buy policy incorrectly formatted"
    buy_policy = buy_policy.set_index("item")

    item_ids = utils.get_item_ids()

    new_snatch = {}
    for item, b in buy_policy.iterrows():
        item_id = item_ids.get(item, None)
        if item_id:
            snatch_item: Dict[str, Any] = {}
            snatch_item["price"] = int(b["buy_price_cap"])
            link_text = f"|cffffffff|Hitem:{item_id}::::::::39:::::::|h[{item}]|h|r"
            snatch_item["link"] = link_text
            new_snatch[f"{item_id}:0:0"] = snatch_item

    return new_snatch


def write_buy_policy() -> None:
    """Writes the buy policy to all accounts."""
    buy_policy = io.reader("outputs", "buy_policy", "parquet")

    cols = ["item", "buy_price_cap"]
    new_snatch = encode_buy_campaign(buy_policy[cols])

    # Read client lua, replace with

    for account in cfg.us.get("accounts", []):
        path = utils.make_lua_path(account_name=account, datasource="Auc-Advanced")
        data = io.reader(name=path, ftype="lua")
        snatch = data["AucAdvancedData"]["UtilSearchUiData"]["Current"]
        snatch["snatch.itemsList"] = {}
        snatch = snatch["snatch.itemsList"]
        data["AucAdvancedData"]["UtilSearchUiData"]["Current"][
            "snatch.itemsList"
        ] = new_snatch
        io.writer(data, name=path, ftype="lua")


def encode_sell_campaign(sell_policy: pd.DataFrame) -> Dict[str, Any]:
    """Encode sell policy dataframe into dictionary."""
    cols = [
        "item",
        "sell_buyout",
        "sell_bid",
        "sell_count",
        "sell_stack",
        "sell_duration",
    ]
    assert (sell_policy.columns == cols).all(), "Sell policy incorrectly formatted"
    sell_policy = sell_policy.set_index("item")

    item_ids = utils.get_item_ids()

    # Seed new appraiser
    new_appraiser: Dict[str, Any] = {
        "bid.markdown": 0,
        "columnsortcurDir": 1,
        "columnsortcurSort": 6,
        "duration": 720,
        "bid.deposit": True,
    }

    for item, d in sell_policy.iterrows():
        code = item_ids[item]

        try:
            new_appraiser[f"item.{code}.fixed.bid"] = int(d["sell_bid"])
        except ValueError:
            raise ValueError(f"{code} for {item} not present")

        new_appraiser[f"item.{code}.fixed.buy"] = int(d["sell_buyout"])
        new_appraiser[f"item.{code}.duration"] = int(d["sell_duration"])
        new_appraiser[f"item.{code}.number"] = int(d["sell_count"])
        new_appraiser[f"item.{code}.stack"] = int(d["sell_stack"])

        new_appraiser[f"item.{code}.bulk"] = True
        new_appraiser[f"item.{code}.match"] = False
        new_appraiser[f"item.{code}.model"] = "fixed"

    return new_appraiser


def write_sell_policy() -> None:
    """Writes the sell policy to accounts."""
    sell_policy = io.reader("outputs", "sell_policy", "parquet")

    cols = [
        "item",
        "sell_buyout",
        "sell_bid",
        "sell_count",
        "sell_stack",
        "sell_duration",
    ]
    new_appraiser = encode_sell_campaign(sell_policy[cols])

    # Read client lua, replace with
    for account in cfg.us.get("accounts", []):
        path = utils.make_lua_path(account_name=account, datasource="Auc-Advanced")
        data = io.reader(name=path, ftype="lua")
        data["AucAdvancedConfig"]["profile.Default"]["util"][
            "appraiser"
        ] = new_appraiser
        io.writer(data, name=path, ftype="lua")


def analyse_sell_policy(
    stack: int = 1,
    max_sell: int = 10,
    duration: str = "m",
    MAX_STD: int = 5,
    MIN_PROFIT: int = 300,
    MIN_PROFIT_PCT: float = 0.015,
) -> None:
    """Creates sell policy based on information."""
    item_table = io.reader("intermediate", "item_table", "parquet")
    listing_each = io.reader("intermediate", "listing_each", "parquet")
    item_volume_change_probability = io.reader(
        "intermediate", "item_volume_change_probability", "parquet"
    )

    cols = [
        "item_deposit",
        "material_make_cost",
        "bbpred_std",
        "bbpred_price",
        "user_max_sell",
        "inv_ahm_bag",
        "replenish_qty",
        "replenish_z",
    ]
    sell_items = item_table[item_table["user_Sell"] == True][cols]
    sell_items["item_deposit"] = sell_items["item_deposit"] * (
        utils.duration_str_to_mins(duration) / (60 * 24)
    )

    sell_items["sell_exp_decay"] = 2 - sell_items["replenish_z"].apply(
        lambda x: norm.cdf(x)
    )

    listing_each = listing_each[listing_each["list_price_z"] < MAX_STD]
    listing_each = listing_each.sort_values(["item", "list_price_per"])
    listing_each["sell_rank"] = (
        listing_each.groupby("item")["list_price_z"].rank(method="first").astype(int)
        - 1
    )

    listing_each = pd.merge(
        item_volume_change_probability,
        listing_each,
        how="left",
        on=["item", "sell_rank"],
    )
    listing_each = listing_each.set_index(["item"])
    listing_each["list_price_z"] = listing_each["list_price_z"].fillna(MAX_STD)

    gouge_price = sell_items["bbpred_price"] + (sell_items["bbpred_std"] * MAX_STD)

    listing_each["list_price_per"] = (
        listing_each["list_price_per"].fillna(gouge_price).astype(int)
    )
    listing_each = listing_each.reset_index().sort_values(["item", "sell_rank"])

    listing_profits = pd.merge(
        listing_each, sell_items, how="left", left_on="item", right_index=True
    )

    listing_profits["sell_buyout"] = listing_profits["list_price_per"] - 9

    listing_profits["sell_estimated_profit"] = (
        (listing_profits["sell_buyout"] * 0.95 - listing_profits["material_make_cost"])
        * (listing_profits["sell_probability"] ** listing_profits["sell_exp_decay"])
    ) - (listing_profits["item_deposit"] * (1 - listing_profits["sell_probability"]))

    best_profits_ind = listing_profits.groupby("item")["sell_estimated_profit"].idxmax()
    sell_policy = listing_profits.loc[best_profits_ind]

    sell_policy["profit_min"] = MIN_PROFIT
    sell_policy["profit_pct"] = MIN_PROFIT_PCT * sell_policy["bbpred_price"]
    sell_policy["profit_feasible"] = sell_policy[["profit_min", "profit_pct"]].max(
        axis=1
    )
    # sell_policy["profit_infeasible"] = (
    #     sell_policy["profit_feasible"] > sell_policy["sell_estimated_profit"]
    # )

    # Shows the amount required to be profitable
    sell_policy["sell_bid"] = (
        sell_policy["sell_buyout"]
        - sell_policy["sell_estimated_profit"]
        + sell_policy["profit_feasible"]
    )

    low_bid_ind = sell_policy[
        sell_policy["sell_bid"] < sell_policy["sell_buyout"]
    ].index
    sell_policy.loc[low_bid_ind, "sell_bid"] = sell_policy.loc[
        low_bid_ind, "sell_buyout"
    ]

    sell_policy["sell_duration"] = utils.duration_str_to_mins(duration)
    sell_policy = sell_policy.sort_values("sell_estimated_profit", ascending=False)

    sell_policy["sell_stack"] = stack
    sell_policy["user_max_sell"] = sell_policy["user_max_sell"].replace(0, max_sell)
    sell_policy["sell_count"] = sell_policy[["inv_ahm_bag", "user_max_sell"]].min(
        axis=1
    )
    sell_policy["sell_count"] = (
        sell_policy["sell_count"] / sell_policy["sell_stack"]
    ).astype(int)

    # TODO Here is likely where we can make it respect min_holding
    sell_policy["sell_min"] = sell_policy[["user_max_sell", "inv_ahm_bag"]].min(axis=1)
    adjust_stack = sell_policy[
        sell_policy["sell_min"] < sell_policy["sell_stack"]
    ].index
    sell_policy.loc[adjust_stack, "sell_stack"] = 1
    sell_policy.loc[adjust_stack, "sell_count"] = sell_policy.loc[
        adjust_stack, "sell_min"
    ]

    io.writer(sell_policy, "outputs", "sell_policy", "parquet")

    listing_profits = listing_profits.set_index(["sell_rank", "item"])[
        "sell_estimated_profit"
    ].unstack()
    io.writer(listing_profits, "reporting", "listing_profits", "parquet")


def analyse_make_policy() -> None:
    """Prints what potions to make."""
    item_table = io.reader("intermediate", "item_table", "parquet")
    item_table.index.name = "item"

    cols = [
        "item_id",
        "user_Make",
        "user_Sell",
        "user_make_pass",
        "user_mean_holding",
        "inv_total_all",
        "inv_ahm_bag",
        "inv_ahm_bank",
    ]
    make_policy = item_table[cols]

    make_policy["make_ideal"] = (
        make_policy["user_mean_holding"] - make_policy["inv_total_all"]
    )
    make_policy["make_counter"] = make_policy["make_ideal"].apply(lambda x: max(x, 0))
    make_policy["make_mat_available"] = (
        make_policy["inv_ahm_bag"] + make_policy["inv_ahm_bank"]
    )
    make_policy["make_actual"] = 0
    make_policy["make_mat_flag"] = 0

    user_items = cfg.ui.copy()

    # Iterates through the table one at a time, to ensure fair distribution of mat usage
    # Tests if reached counter and is made from stuff
    # Checks the material count can go down first before decrementing
    # If after each check, append to list to see for any changes on any pass through
    change = [True]
    while any(change):
        change = []

        for item, row in make_policy.iterrows():

            made_from = user_items[item].get("made_from", {})
            under_counter = row["make_actual"] < row["make_counter"]
            user_make_pass = row["user_make_pass"]

            if made_from and under_counter and not (user_make_pass):
                item_increment = True
                for material, qty in made_from.items():
                    if "Vial" not in material:
                        item_increment = (
                            make_policy.loc[material, "make_mat_available"] >= qty
                        ) & item_increment

                if item_increment:
                    for material, qty in user_items[item].get("made_from", {}).items():
                        make_policy.loc[material, "make_mat_available"] -= qty
                        make_policy.loc[material, "make_mat_flag"] = 1
                    make_policy.loc[item, "make_actual"] += 1

                change.append(item_increment)

    io.writer(make_policy, "outputs", "make_policy", "parquet")


def encode_make_policy(
    make_policy: pd.DataFrame,
) -> Tuple[Dict[str, int], Dict[int, str]]:
    """Encodes make campaign dataframe into dictionary."""
    make_policy = io.reader("outputs", "make_policy", "parquet")

    new_craft_queue = make_policy[
        (make_policy["user_make_pass"] == 0) & (make_policy["make_actual"] > 0)
    ]["make_actual"].to_dict()

    # Ordering important here for overwrites
    make_policy["group"] = "Other"
    make_policy.loc[make_policy[make_policy["user_Sell"] == 1].index, "group"] = "Sell"
    make_policy.loc[
        make_policy[make_policy["make_mat_flag"] == 1].index, "group"
    ] = "Materials"
    make_policy = make_policy[make_policy["item_id"] > 0]

    item_groups = make_policy.set_index(["item_id"])["group"].to_dict()

    return new_craft_queue, item_groups


def write_make_policy() -> None:
    """Writes the make policy to all accounts."""
    make_policy = io.reader("outputs", "make_policy", "parquet")
    new_craft_queue, item_groups = encode_make_policy(make_policy)

    ahm = utils.get_ahm()
    path = utils.make_lua_path(
        account_name=ahm["account"], datasource="TradeSkillMaster"
    )
    content = io.reader(name=path, ftype="lua", custom="rb")

    craft_mark = f'f@Alliance - {cfg.us["server"]}@internalData@crafts'
    start, end = utils.find_tsm_marker(content, f'["{craft_mark}"]'.encode("ascii"))

    crafting_dict = lua.decode("{" + content[start:end].decode("ascii") + "}")
    for _, item_data in crafting_dict[craft_mark].items():
        item_name = item_data.get("name", "_no_name")
        queued = new_craft_queue.get(item_name, 0)
        if "queued" in item_data:
            item_data["queued"] = queued

    new_craft = utils.dict_to_lua(crafting_dict).encode("ascii")
    new_craft = new_craft.replace(
        f"\n{craft_mark}".encode("ascii"), f'\n["{craft_mark}"]'.encode("ascii"),
    )
    content = content[:start] + new_craft + content[end:]

    # Update item groups
    groups_mark = '["p@Default@userData@items"]'
    item_text = f"{groups_mark} = " + "{"
    for item_code, group in item_groups.items():
        item_text += f'["i:{item_code}"] = "{group}", '
    item_text += "}"
    start, end = utils.find_tsm_marker(content, groups_mark.encode("ascii"))
    content = content[:start] + item_text.encode("ascii") + content[end:]

    io.writer(content, name=path, ftype="lua", custom="wb")



"""Main entry point for pricer program.

First time running checklist
- Set up TSM groups
- Run booty bay
- Get icons

"""
import argparse
from datetime import datetime as dt
import logging
import warnings

from tqdm import tqdm

from . import analysis, campaign, config as cfg, reporting, sources
from .webserver.views import app


warnings.simplefilter(action="ignore")
logger = logging.getLogger(__name__)


def run_analytics(
    stack: int = 5, max_sell: int = 20, duration: str = "m", test: bool = False
) -> None:
    """Run the main analytics pipeline."""
    with tqdm(total=220) as pbar:
        run_dt = dt.now().replace(microsecond=0)
        # TODO remove this run_dt crap
        if not test:
            sources.get_arkinventory_data()
            sources.get_beancounter_data()
            sources.get_auctioneer_data()
        pbar.update(10)

        sources.clean_bb_data()
        pbar.update(10)

        sources.clean_arkinventory_data(run_dt)
        pbar.update(10)

        sources.clean_beancounter_data()
        pbar.update(10)

        sources.clean_auctioneer_data()
        pbar.update(10)

        sources.clean_item_skeleton()
        pbar.update(10)

        analysis.create_item_inventory()
        pbar.update(10)

        analysis.create_item_facts()
        pbar.update(10)

        analysis.predict_item_prices()
        pbar.update(10)

        analysis.analyse_rolling_buyout()
        pbar.update(10)

        analysis.analyse_material_cost()
        pbar.update(10)

        analysis.analyse_listings()
        pbar.update(10)

        analysis.analyse_replenishment()
        pbar.update(10)

        analysis.merge_item_table()
        pbar.update(10)

        analysis.predict_volume_sell_probability(duration)
        pbar.update(10)

        campaign.analyse_buy_policy()
        pbar.update(10)

        campaign.write_buy_policy()
        pbar.update(10)

        campaign.analyse_sell_policy(stack=stack, max_sell=max_sell, duration=duration)
        pbar.update(10)

        campaign.write_sell_policy()
        pbar.update(10)

        campaign.analyse_make_policy()
        pbar.update(10)

        campaign.write_make_policy()
        pbar.update(10)


def run_reporting() -> None:
    """Run steps to create plots and insights."""
    analysis.report_profits()

    reporting.have_in_bag()
    reporting.make_missing()
    reporting.produce_item_reporting()
    reporting.produce_listing_items()
    reporting.produce_activity_tracking()
    reporting.draw_profit_charts()


def main() -> None:
    """Main program runner."""
    run_dt = dt.now().replace(microsecond=0)

    parser = argparse.ArgumentParser(description="WoW Auctions")
    parser.add_argument("-b", help="Update web booty bay analysis", action="store_true")
    parser.add_argument("-i", help="Get item icons for webserver", action="store_true")

    parser.add_argument("-s", type=int, default=5, help="Stack size")
    parser.add_argument("-m", type=int, default=20, help="Max sell")
    parser.add_argument("-d", type=str, default="m", help="Duration")

    parser.add_argument("-f", help="Start flask webserver", action="store_true")
    parser.add_argument("-n", help="No analysis, skip", action="store_true")

    parser.add_argument("-r", help="Generate reporting and plots", action="store_true")

    parser.add_argument("-t", help="Run on test data", action="store_true")

    parser.add_argument("-v", help="Verbose mode (info)", action="store_true")
    parser.add_argument("-vv", help="Verbose mode (debug)", action="store_true")
    args = parser.parse_args()

    cfg.set_loggers(base_logger=logger, v=args.v, vv=args.vv)
    logger.info("Program started, arguments parsed")
    logger.debug(args)

    if args.b:
        sources.get_bb_data()
    if args.i:
        sources.get_item_icons()
    if args.t:
        """Test environment."""
        cfg.env = {"basepath": "data/_test"}
        test_items = ["Mighty Rage Potion", "Gromsblood", "Crystal Vial"]
        cfg.ui = {k: v for k, v in cfg.ui.items() if k in test_items}

    if not args.n:
        run_analytics(stack=args.s, max_sell=args.m, duration=args.d, test=args.t)

    if args.r:
        run_reporting()

    logger.info(f"Program end, seconds {(dt.now() - run_dt).total_seconds()}")

    if args.f:
        logger.info("Starting webserver")
        app.run(debug=True, threaded=True)


if __name__ == "__main__":
    main()



"""Analyses cleaned data sources to form intermediate tables."""
import logging
from typing import Any, Dict

from numpy import inf
import pandas as pd
from scipy.stats import gaussian_kde

from pricer import config as cfg, io, utils

logger = logging.getLogger(__name__)


def predict_item_prices() -> None:
    """Analyse exponential average mean and std of items given 14 day, 2 hour history."""
    bb_fortnight = io.reader("cleaned", "bb_fortnight", "parquet")
    user_items = cfg.ui.copy()

    predicted_prices = _predict_item_prices(bb_fortnight, user_items)
    io.writer(predicted_prices, "intermediate", "predicted_prices", "parquet")


def _predict_item_prices(
    bb_fortnight: pd.DataFrame, user_items: Dict[str, Any]
) -> pd.DataFrame:
    # Work out if an item is auctionable, or get default price
    item_prices = pd.DataFrame()
    for item_name, item_details in user_items.items():
        user_vendor_price = item_details.get("vendor_price")

        if user_vendor_price:
            item_prices.loc[item_name, "bbpred_price"] = user_vendor_price
            item_prices.loc[item_name, "bbpred_std"] = 0
        else:
            q = cfg.us["analysis"]["ITEM_PRICE_OUTLIER_CAP"]
            df = bb_fortnight[bb_fortnight["item"] == item_name]
            df["silver"] = df["silver"].clip(
                lower=df["silver"].quantile(q), upper=df["silver"].quantile(1 - q),
            )
            try:
                item_prices.loc[item_name, "bbpred_price"] = int(
                    df["silver"].ewm(alpha=0.2).mean().iloc[-1]
                )
                item_prices.loc[item_name, "bbpred_std"] = (
                    df["silver"].std().astype(int)
                )
            except IndexError:
                logging.exception(
                    f"""Price prediction problem for {item_name}.
                    Did you add something and not use booty bay?"""
                )

    qty_df = bb_fortnight[bb_fortnight["snapshot"] == bb_fortnight["snapshot"].max()]
    qty_df = qty_df.set_index("item")["quantity"]
    qty_df.name = "bbpred_quantity"

    predicted_prices = item_prices.join(qty_df).fillna(0).astype(int)
    return predicted_prices


def analyse_rolling_buyout() -> None:
    """Builds rolling average of user's auction purchases using beancounter data."""
    bean_purchases = io.reader("cleaned", "bean_purchases", "parquet")

    bean_buys = bean_purchases["item"].isin(utils.user_item_filter("Buy"))
    bean_purchases = bean_purchases[bean_buys].sort_values(["item", "timestamp"])

    cols = ["item", "buyout_per"]
    purchase_each = utils.enumerate_quantities(bean_purchases, cols=cols, qty_col="qty")

    # Needed to ensure that groupby will work for a single item
    purchase_each.loc[purchase_each.index.max() + 1] = ("dummy", 0)

    SPAN = cfg.us["analysis"]["ROLLING_BUYOUT_SPAN"]
    ewm = (
        purchase_each.groupby("item")
        .apply(lambda x: x["buyout_per"].ewm(span=SPAN).mean())
        .reset_index()
    )
    # Get the latest item value
    latest_item = ewm.groupby("item")["level_1"].max()
    bean_rolling_buyout = ewm.loc[latest_item].set_index("item")[["buyout_per"]]

    bean_rolling_buyout = bean_rolling_buyout.drop("dummy").astype(int)
    bean_rolling_buyout.columns = ["bean_rolling_buyout"]
    io.writer(bean_rolling_buyout, "intermediate", "bean_rolling_buyout", "parquet")


def analyse_material_cost() -> None:
    """Analyse cost of materials for items, using purchase history or BB predicted price."""
    bean_rolling_buyout = io.reader("intermediate", "bean_rolling_buyout", "parquet")
    item_prices = io.reader("intermediate", "predicted_prices", "parquet")
    mat_prices = item_prices.join(bean_rolling_buyout)

    r = cfg.us["analysis"]["BB_MAT_PRICE_RATIO"]

    # Material costs are taken as a ratio of booty bay prices, and (recent) actual buyouts
    mat_prices["material_buyout_cost"] = (
        mat_prices["bean_rolling_buyout"].fillna(mat_prices["bbpred_price"]) * (1 - r)
        + (mat_prices["bbpred_price"] * r)
    ).astype(int)

    mat_prices["material_make_cost"] = 0

    # Determine raw material cost for manufactured items
    for item_name, item_details in cfg.ui.items():
        material_cost = 0
        user_made_from = item_details.get("made_from", {})
        if user_made_from:
            for ingredient, count in user_made_from.items():
                material_cost += (
                    mat_prices.loc[ingredient, "material_buyout_cost"] * count
                )
        else:
            material_cost = mat_prices.loc[item_name, "material_buyout_cost"]
        mat_prices.loc[item_name, "material_make_cost"] = int(material_cost)

    mat_prices = mat_prices[["material_buyout_cost", "material_make_cost"]]
    io.writer(mat_prices, "intermediate", "mat_prices", "parquet")


def create_item_inventory() -> None:
    """Convert Arkinventory tabular data into dataframe of counts for user items."""
    item_inventory = io.reader("cleaned", "ark_inventory", "parquet")

    roles = {char["name"]: char["role"] for char in cfg.us.get("roles", {})}

    item_inventory["role"] = item_inventory["character"].apply(
        lambda x: roles[x] if x in roles else "char"
    )
    role_types = ["ahm", "mule", "char"]
    assert item_inventory["role"].isin(role_types).all()

    location_rename = {
        "Inventory": "bag",
        "Bank": "bank",
        "Auctions": "auc",
        "Mailbox": "mail",
    }
    item_inventory["loc_short"] = item_inventory["location"].replace(location_rename)
    item_inventory["inv"] = (
        "inv_" + item_inventory["role"] + "_" + item_inventory["loc_short"]
    )

    item_inventory = item_inventory.groupby(["inv", "item"]).sum()["count"].unstack().T

    # Ensure 9x grid of columns
    for role in role_types:
        for loc in location_rename.values():
            col = f"inv_{role}_{loc}"
            if col not in item_inventory.columns:
                item_inventory[col] = 0

    item_inventory = item_inventory.fillna(0).astype(int)

    # Analyse aggregate; ordering important here
    item_inventory["inv_total_all"] = item_inventory.sum(axis=1)

    cols = [x for x in item_inventory.columns if "ahm" in x or "mule" in x]
    item_inventory["inv_total_hold"] = item_inventory[cols].sum(axis=1)

    cols = [x for x in item_inventory.columns if "ahm" in x]
    item_inventory["inv_total_ahm"] = item_inventory[cols].sum(axis=1)

    io.writer(item_inventory, "intermediate", "item_inventory", "parquet")


def analyse_replenishment() -> None:
    """Determine the demand for item replenishment."""
    item_skeleton = io.reader("cleaned", "item_skeleton", "parquet")
    item_inventory = io.reader("intermediate", "item_inventory", "parquet")

    replenish = item_skeleton.join(item_inventory).fillna(0).astype(int)

    user_items = cfg.ui.copy()

    replenish["replenish_qty"] = (
        replenish["user_mean_holding"] - replenish["inv_total_all"]
    )

    # Update replenish list with user_made_from
    for item, row in replenish.iterrows():
        if row["replenish_qty"] > 0:
            for ingredient, count in user_items[item].get("made_from", {}).items():
                replenish.loc[ingredient, "replenish_qty"] += (
                    count * row["replenish_qty"]
                )

    replenish["replenish_z"] = (
        replenish["replenish_qty"] / replenish["user_std_holding"]
    )
    replenish["replenish_z"] = (
        replenish["replenish_z"].replace([inf, -inf], 0).fillna(0)
    )

    replenish = replenish[["replenish_qty", "replenish_z"]]
    io.writer(replenish, "intermediate", "replenish", "parquet")


def create_item_facts() -> None:
    """Collate simple item facts."""
    item_skeleton = io.reader("cleaned", "item_skeleton", "parquet")
    bb_deposit = io.reader("cleaned", "bb_deposit", "parquet")
    item_ids = utils.get_item_ids()

    item_facts = item_skeleton.join(bb_deposit)[["item_deposit"]].join(
        pd.Series(item_ids, name="item_id")
    )
    item_facts = item_facts.fillna(0).astype(int)

    io.writer(item_facts, "cleaned", "item_facts", "parquet")


def merge_item_table() -> None:
    """Combine item information into single master table."""
    item_skeleton = io.reader("cleaned", "item_skeleton", "parquet")
    mat_prices = io.reader("intermediate", "mat_prices", "parquet")
    item_facts = io.reader("cleaned", "item_facts", "parquet")
    item_inventory = io.reader("intermediate", "item_inventory", "parquet")
    predicted_prices = io.reader("intermediate", "predicted_prices", "parquet")
    replenish = io.reader("intermediate", "replenish", "parquet")

    item_table = (
        item_skeleton.join(mat_prices)
        .join(predicted_prices)
        .join(item_inventory)
        .join(item_facts)
        .join(replenish)
    ).fillna(0)

    io.writer(item_table, "intermediate", "item_table", "parquet")


def analyse_listings() -> None:
    """Convert live listings into single items."""
    auc_listings = io.reader("cleaned", "auc_listings", "parquet")
    auc_listings = auc_listings[auc_listings["item"].isin(cfg.ui)]

    predicted_prices = io.reader("intermediate", "predicted_prices", "parquet")

    ranges = pd.merge(
        auc_listings,
        predicted_prices,
        how="left",
        left_on="item",
        right_index=True,
        validate="m:1",
    )
    ranges["price_z"] = (ranges["price_per"] - ranges["bbpred_price"]) / ranges[
        "bbpred_std"
    ]

    cols = ["item", "price_per", "price_z"]
    listing_each = utils.enumerate_quantities(ranges, cols=cols)
    listing_each.columns = ["item", "list_price_per", "list_price_z"]

    io.writer(listing_each, "intermediate", "listing_each", "parquet")


def predict_volume_sell_probability(dur_char: str = "m") -> None:
    """Expected volume changes as a probability of sale given BB recent history."""
    bb_fortnight = io.reader("cleaned", "bb_fortnight", "parquet")
    user_sells = utils.user_item_filter("Sell")
    MAX_LISTINGS = cfg.us["analysis"]["MAX_LISTINGS_PROBABILITY"]

    duration_mins = utils.duration_str_to_mins(dur_char)
    polls = int(duration_mins / 60 / 2)
    logger.debug(f"Analysing volume sell prob based on {polls} snapshot periods")

    item_volume_change_probability = pd.DataFrame(columns=user_sells)
    for item in user_sells:
        item_fortnight = bb_fortnight[bb_fortnight["item"] == item]

        volume_df = pd.DataFrame()
        price_df = pd.DataFrame()
        for i in range(1, polls + 1):
            item_fortnight["snapshot_prev"] = item_fortnight["snapshot"].shift(i)
            offset_test = pd.merge(
                item_fortnight,
                item_fortnight,
                left_on=["snapshot", "item"],
                right_on=["snapshot_prev", "item"],
            )
            volume_df[i] = offset_test["quantity_y"] - offset_test["quantity_x"]
            price_df[i] = offset_test["silver_y"] - offset_test["silver_x"]

        volume_df = volume_df.dropna().mean(axis=1)
        price_df = price_df.dropna().mean(axis=1)
        volume_df = volume_df[price_df <= 0]
        try:
            gkde = gaussian_kde(volume_df)
        except ValueError as e:
            raise ValueError(
                f"Could not analyse {item}, is this new and needs bb?"
            ) from e

        listing_range = range(-MAX_LISTINGS + 1, 1)
        probability = pd.Series(gkde(listing_range), index=listing_range)

        probability = probability.cumsum()
        probability.index = [-i for i in probability.index]

        item_volume_change_probability[item] = probability.sort_index()

    item_volume_change_probability.index.name = "sell_rank"
    item_volume_change_probability.columns.name = "item"
    item_volume_change_probability = item_volume_change_probability.stack()
    item_volume_change_probability.name = "sell_probability"
    item_volume_change_probability = item_volume_change_probability.reset_index()

    io.writer(
        item_volume_change_probability,
        "intermediate",
        "item_volume_change_probability",
        "parquet",
    )


def report_profits() -> None:
    """Compare purchases and sales to expected value to derive profit from action."""
    bean_results = io.reader("cleaned", "bean_results", "parquet")
    bean_purchases = io.reader("cleaned", "bean_purchases", "parquet")
    bb_history = io.reader("cleaned", "bb_history", "parquet")

    bean_results["date"] = bean_results["timestamp"].dt.date.astype("datetime64")
    bean_results["profit"] = bean_results["received"].fillna(
        -bean_results["item_deposit"]
    )
    bean_results["qty_change"] = bean_results.apply(
        lambda x: -x["qty"] if x["auction_type"] == "completedAuctions" else 0, axis=1
    )

    bean_purchases["date"] = bean_purchases["timestamp"].dt.date.astype("datetime64")
    bean_purchases["qty_change"] = bean_purchases["qty"]
    bean_purchases["profit"] = -bean_purchases["buyout"]

    purchase_change = bean_purchases.groupby(["item", "date"])[
        ["qty_change", "profit"]
    ].sum()
    purchase_change.columns = ["purchase_qty_change", "purchase_profit"]

    result_change = bean_results.groupby(["auction_type", "item", "date"])[
        ["qty_change", "profit"]
    ].sum()
    completed_change = result_change.loc["completedAuctions"]
    completed_change.columns = ["completed_qty_change", "completed_profit"]
    failed_change = result_change.loc["failedAuctions"]
    failed_change.columns = ["failed_qty_change", "failed_profit"]

    bb_history = bb_history[bb_history["date"] >= bean_results["date"].min()]
    bb_history = bb_history.set_index(["item", "date"])

    profits = (
        bb_history.join(purchase_change)
        .join(completed_change)
        .join(failed_change)
        .fillna(0)
        .astype(int)
    )

    profits["total_action"] = profits[[x for x in profits if "_profit" in x]].sum(
        axis=1
    )
    profits["total_qty"] = profits[[x for x in profits if "_qty_change" in x]].sum(
        axis=1
    )

    # vector style material cost calculation
    material_update = []
    for item_name, item_details in cfg.ui.items():
        if item_name in profits.index:
            material_cost = pd.Series(
                0, index=profits.loc[item_name].index, name="silveravg"
            )
            user_made_from = item_details.get("made_from", {})
            if user_made_from:
                for ingredient, count in user_made_from.items():
                    if ingredient in profits.index:
                        material_cost += profits.loc[ingredient, "silveravg"] * count
                    else:
                        material_cost += cfg.ui[ingredient]["vendor_price"] * count
            else:
                material_cost = profits.loc[item_name, "silveravg"]
            material_cost = material_cost.reset_index()
            material_cost["item"] = item_name
            material_update.append(material_cost)

    material_updates = pd.concat(material_update)

    profits = profits.join(
        material_updates.set_index(["item", "date"])["silveravg"], rsuffix="_cost"
    )
    profits["total_materials"] = -profits["silveravg_cost"] * profits["total_qty"]
    profits["total_profit"] = profits["total_action"] - profits["total_materials"]

    io.writer(profits, "reporting", "profits", "parquet")



"""Schema enforcement for project."""
import pandera as pa
from pandera import Column, Index


ark_inventory_schema = pa.DataFrameSchema(
    columns={
        "character": Column(pa.String),
        "location": Column(pa.String),
        "item": Column(pa.String),
        "count": Column(pa.Int),
        "timestamp": Column(pa.DateTime),
    },
    strict=True,
)

ark_monies_schema = pa.DataFrameSchema(
    columns={"monies": Column(pa.Int), "timestamp": Column(pa.DateTime)},
    strict=True,
    index=Index(pa.String, name="character"),
)

auc_listings_raw_schema = pa.DataFrameSchema(
    columns={
        0: Column(pa.String, nullable=True),
        1: Column(pa.String, nullable=True),
        2: Column(pa.String, nullable=True),
        3: Column(pa.String, nullable=True),
        4: Column(pa.String, nullable=True),
        5: Column(pa.String, nullable=True),
        6: Column(pa.String, nullable=True),
        7: Column(pa.String, nullable=True),
        8: Column(pa.String, nullable=True),
        9: Column(pa.String, nullable=True),
        10: Column(pa.String, nullable=True),
        11: Column(pa.String, nullable=True),
        12: Column(pa.String, nullable=True),
        13: Column(pa.String, nullable=True),
        14: Column(pa.String, nullable=True),
        15: Column(pa.String, nullable=True),
        16: Column(pa.String, nullable=True),
        17: Column(pa.String, nullable=True),
        18: Column(pa.String, nullable=True),
        19: Column(pa.String, nullable=True),
        20: Column(pa.String, nullable=True),
        21: Column(pa.String, nullable=True),
        22: Column(pa.String, nullable=True),
        23: Column(pa.String, nullable=True),
        24: Column(pa.String, nullable=True),
        25: Column(pa.String, nullable=True),
        26: Column(pa.String, nullable=True),
        27: Column(pa.String, nullable=True),
    }
)

auc_listings_schema = pa.DataFrameSchema(
    columns={
        "item": Column(pa.String),
        "quantity": Column(pa.Int),
        "buy": Column(pa.Int),
        "sellername": Column(pa.String),
        "price_per": Column(pa.Int),
        "time_remaining": Column(pa.Int),
    },
    strict=True,
)

bb_alltime_schema = pa.DataFrameSchema(
    columns={
        "date": Column(pa.DateTime),
        "silver": Column(pa.Int),
        "quantity": Column(pa.Int),
        "item": Column(pa.String),
    },
    strict=True,
    index=Index(pa.Int),
)

bb_deposit_schema = pa.DataFrameSchema(
    columns={"item_deposit": Column(pa.Int, nullable=True)},
    strict=True,
    index=Index(pa.String, name="item"),
)

bb_history_schema = pa.DataFrameSchema(
    columns={
        "date": Column(pa.DateTime),
        "silvermin": Column(pa.Int),
        "silveravg": Column(pa.Int),
        "silvermax": Column(pa.Int),
        "silverstart": Column(pa.Int),
        "silverend": Column(pa.Int),
        "quantitymin": Column(pa.Int),
        "quantityavg": Column(pa.Int),
        "quantitymax": Column(pa.Int),
        "item": Column(pa.String),
    },
    strict=True,
    index=Index(pa.Int),
)

bb_fortnight_schema = pa.DataFrameSchema(
    columns={
        "snapshot": Column(pa.DateTime),
        "silver": Column(pa.Int),
        "quantity": Column(pa.Int),
        "item": Column(pa.String),
    },
    strict=True,
    index=Index(pa.Int),
)

beancounter_raw_schema = pa.DataFrameSchema(
    columns={
        0: Column(pa.String, nullable=True),
        1: Column(pa.String, nullable=True),
        2: Column(pa.String, nullable=True),
        3: Column(pa.String, nullable=True),
        4: Column(pa.String, nullable=True),
        5: Column(pa.String, nullable=True),
        6: Column(pa.String, nullable=True),
        7: Column(pa.String, nullable=True),
        8: Column(pa.String, nullable=True),
        9: Column(pa.String, nullable=True),
        10: Column(pa.String, nullable=True),
        11: Column(pa.String, nullable=True),
        12: Column(pa.String, nullable=True),
    },
    strict=True,
)

bean_purchases_schema = pa.DataFrameSchema(
    columns={
        "auction_type": Column(pa.String),
        "item": Column(pa.String),
        "buyer": Column(pa.String),
        "qty": Column(pa.Int),
        "buyout": Column(pa.Int, nullable=True),
        "bid": Column(pa.Int),
        "seller": Column(pa.String),
        "timestamp": Column(pa.DateTime),
        "buyout_per": Column(pa.Float),
        "bid_per": Column(pa.Float),
    },
    strict=True,
)

bean_posted_schema = pa.DataFrameSchema(
    columns={
        "auction_type": Column(pa.String),
        "item": Column(pa.String),
        "seller": Column(pa.String),
        "qty": Column(pa.Int),
        "item_deposit": Column(pa.Int),
        "buyout": Column(pa.Int, nullable=True),
        "bid": Column(pa.Int),
        "timestamp": Column(pa.DateTime),
        "buyout_per": Column(pa.Float),
        "bid_per": Column(pa.Float),
    },
    strict=True,
)

bean_failed_schema = pa.DataFrameSchema(
    columns={
        "auction_type": Column(pa.String),
        "item": Column(pa.String),
        "seller": Column(pa.String),
        "qty": Column(pa.Int),
        "item_deposit": Column(pa.Int),
        "buyout": Column(pa.Int, nullable=True),
        "bid": Column(pa.Int),
        "timestamp": Column(pa.DateTime),
        "buyout_per": Column(pa.Float),
        "bid_per": Column(pa.Float),
    },
    strict=True,
)

bean_success_schema = pa.DataFrameSchema(
    columns={
        "auction_type": Column(pa.String),
        "item": Column(pa.String),
        "seller": Column(pa.String),
        "qty": Column(pa.Int),
        "received": Column(pa.Int),
        "item_deposit": Column(pa.Int),
        "ah_cut": Column(pa.Int),
        "buyout": Column(pa.Int, nullable=True),
        "bid": Column(pa.Int),
        "buyer": Column(pa.String),
        "timestamp": Column(pa.DateTime),
        "received_per": Column(pa.Float),
        "buyout_per": Column(pa.Float),
        "bid_per": Column(pa.Float),
    },
    strict=True,
)

bean_results_schema = pa.DataFrameSchema(
    columns={
        "auction_type": Column(pa.String),
        "item": Column(pa.String),
        "seller": Column(pa.String),
        "qty": Column(pa.Int),
        "received": Column(pa.Int, nullable=True),
        "item_deposit": Column(pa.Int),
        "ah_cut": Column(pa.Int, nullable=True),
        "buyout": Column(pa.Int, nullable=True),
        "bid": Column(pa.Int),
        "buyer": Column(pa.String, nullable=True),
        "timestamp": Column(pa.DateTime),
        "received_per": Column(pa.Float, nullable=True),
        "buyout_per": Column(pa.Float),
        "bid_per": Column(pa.Float),
        "success": pa.Column(pa.Int),
    },
    strict=True,
)

item_skeleton_raw_schema = pa.DataFrameSchema(
    columns={
        "user_min_holding": Column(pa.Int, nullable=True),
        "user_max_holding": Column(pa.Int, nullable=True),
        "user_max_sell": Column(pa.Int, nullable=True),
        "user_Buy": Column(nullable=True),
        "user_Sell": Column(nullable=True),
        "user_Make": Column(nullable=True),
        "user_made_from": Column(pa.Object, nullable=True),
        "user_make_pass": Column(pa.Int, nullable=True),
        "user_vendor_price": Column(pa.Int, nullable=True),
    },
    strict=True,
    index=Index(pa.String),
)

item_skeleton_schema = pa.DataFrameSchema(
    columns={
        "user_min_holding": Column(pa.Int),
        "user_max_holding": Column(pa.Int),
        "user_max_sell": Column(pa.Int, nullable=True),
        "user_Buy": Column(pa.Int),
        "user_Sell": Column(pa.Int),
        "user_Make": Column(pa.Int),
        "user_make_pass": Column(pa.Int),
        "user_vendor_price": Column(pa.Int),
        "user_std_holding": Column(pa.Float),
        "user_mean_holding": Column(pa.Int),
    },
    strict=True,
    index=Index(pa.String),
)
